package OJS.Midi;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiDevice;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Synthesizer;
import javax.sound.midi.Sequencer;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.Receiver;
import javax.sound.midi.Transmitter;

/**
 * @author shengz
 *
 * This class instance is a representation of the midi input port in software. 
 * It listens to the input port and transmits messages whenever it receives them. 
 */
public class MidiIn 
	implements Transmitter
{
	// MidiDevice to represent MIDI IN port
	MidiDevice	device = null;
	
	MidiDevice.Info[] deviceInfo = null;
	
	// Transmitter owned by
	Transmitter t = null; 

	/**
	 * Seach for the midi input port and instantiate a device for it.
	 */
	public MidiIn()
	{
		
		// Obtain array of Midi Devices
		deviceInfo = MidiSystem.getMidiDeviceInfo();
		if (deviceInfo == null)
		{
			out("wasn't able to retrieve MidiDevice info");
			System.exit(1);
		}

		out("obtained midi device list");
		
		// Enumerate the devices to find MIDI IN Port
		for (int i = 0; i < deviceInfo.length; i++)
		{
			try
			{
				device = MidiSystem.getMidiDevice(deviceInfo[i]);
				
				// Detect whether the device is a hardware port
				if ( ! ( device instanceof Sequencer) && ! (device instanceof Synthesizer)) 
				{
					// out("found MIDI port " + deviceInfo[i].getName());
				
					// We're now sure that device represents a MIDI port
					// Detect the MIDI IN port
					if(device.getMaxTransmitters() != 0)
					{
						
						// MIDI IN port found
						out("MIDI IN port " + deviceInfo[i].getName());						
												
						if (device.isOpen())
						{
							out("device is already open");
							t = device.getTransmitter();
							break;
						} 
						else 
						{
							//out("open MIDI IN port");
							device.open();				

							t = device.getTransmitter();	
							//out("opened MIDI IN port");
							break;
						}						
					}
			    }
				device = null;
			}
			catch (MidiUnavailableException e)
			{
			}			
		}
		if (device == null) 
		{
			out("No MIDI input device found");
		}
	}

	/* Set the receiver where the messages is to be transmitted.
	 * @see javax.sound.midi.Transmitter#setReceiver(javax.sound.midi.Receiver)
	 */
	public void setReceiver(Receiver r)
	{
		if(t != null)
			t.setReceiver(r);
	}
	
	/* (non-Javadoc)
	 * @see javax.sound.midi.Transmitter#getReceiver()
	 */
	public Receiver getReceiver()
	{
		return t.getReceiver();
	}
	
	public void close()
	{
		t.close();
		device.close();
	}
	
	private static void out(String strMessage)
	{
		System.out.println(strMessage);
	}

}

