package OJS.Midi;

import java.io.File;

import javax.sound.midi.Instrument;
import javax.sound.midi.InvalidMidiDataException;
import javax.sound.midi.MidiDevice;
import javax.sound.midi.MidiSystem;
import javax.sound.midi.MidiUnavailableException;
import javax.sound.midi.Sequencer;
import javax.sound.midi.Synthesizer;
import javax.sound.midi.ShortMessage;
import javax.sound.midi.MidiMessage;
import javax.sound.midi.Receiver;


/**
 * @author shengz
 *
 * This class instance is a representation of the midi output device in software. 
 * It provides methods for accessing the midi device. 
 */

public class MidiOut 
	implements Receiver
{
	
	Synthesizer synth = null;
	Receiver r = null;	
	int LOCAL_CHANNEL = 1;
	int REMOTE_CHANNEL = 2;
	
	// MidiDevice to represent MIDI IN port
	MidiDevice	device = null;
	
	MidiDevice.Info[] deviceInfo = null;	
	
		

	/**
	 * @param useDefaultSynth - whether to use the Java synthesizer
	 */
	public MidiOut(boolean useDefaultSynth)
	{
		// Select synthesizer
		
		if (useDefaultSynth){
			try
			{
				synth = MidiSystem.getSynthesizer(); //get default synthsizer
				synth.open();
	        	Instrument[] instr = synth.getAvailableInstruments();
	           	if(instr == null)
	           	{
	           		instr = MidiSystem.getSoundbank( new File("soundbank-min.gm")).getInstruments();
	           	   	System.out.println("Loading default sound bank");
	           	}			
				r = synth.getReceiver();
				out("MIDI Synthesizer: " + synth.getDeviceInfo().getName());										
			}
			catch (Exception e)
			{
			}					
		} 
		else
		{
			// Obtain array of Midi Devices
			deviceInfo = MidiSystem.getMidiDeviceInfo();
			if (deviceInfo == null)
			{
				out("wasn't able to retrieve MidiDevice info");
				System.exit(1);
			}

			out("obtained midi device list");
			
			// Enumerate the devices to find Synthesizer
			for (int i = 0; i < deviceInfo.length; i++)
			{
				try
				{   
					if (-1 != deviceInfo[i].getName().lastIndexOf("Microsoft GS Wavetable")) //TODO a better way to find synthesizer
					{
						device = MidiSystem.getMidiDevice(deviceInfo[i]);
						
						out("MIDI Synthesizer: " + deviceInfo[i].getName());						
												
						if (device.isOpen())
						{
							out("device is already open");
							r = device.getReceiver();
							break;
						} 
						else 
						{
							device.open();				

							r = device.getReceiver();	
							break;
						}							
					}																						   
				}
				catch (MidiUnavailableException e)
				{
				}			
			}
		}		
	}
	
	
	public void close()
	{
		if(r != null) r.close();
		if(synth != null) synth.close();
		if(device != null) device.close();
	}
	
	/** 
	 * Send midi message to synthesizer. Messege's channel is over written to 
	 * the local channel value.
	 * @see javax.sound.midi.Receiver#send(javax.sound.midi.MidiMessage, long)
	 */
	public void send(MidiMessage message, long t)
	{	
		if(message instanceof ShortMessage) {
			// Ignore system messages 
			
			if( (((ShortMessage)message).getCommand() != 0xF0))
			{	
				try{
					ShortMessage s = new ShortMessage();
					s.setMessage(((ShortMessage)message).getCommand(), 
								LOCAL_CHANNEL, 
								((ShortMessage)message).getData1(),
								((ShortMessage)message).getData2());
					
					r.send(s, t);
					
				} catch (InvalidMidiDataException e){}
			}		
		}
//		if ( t < 0 ){
//			r.send(message, -1);		
//		}
//		else 
//		{
//			try
//			{
//				Thread.sleep(t); // wait for the delay
//			}
//			catch (InterruptedException e)
//			{
//			}
//			r.send(message, -1);	
//		}
		
	}
	
	/**
	 * Send the midi messages that are received from the network to the synthesizer.
	 * Messege's channel is over written to the remote channel value.
	 * @param message - message to be sent
	 * @param t - delay
	 */
	public void SendRemote(MidiMessage message, long t){
		if(message instanceof ShortMessage) {
			// Ignore system messages 
			
			if( (((ShortMessage)message).getCommand() != 0xF0))
			{	
				try{
					ShortMessage s = new ShortMessage();
					s.setMessage(((ShortMessage)message).getCommand(), 
								REMOTE_CHANNEL, 
								((ShortMessage)message).getData1(),
								((ShortMessage)message).getData2());
					
					r.send(s, t);
					
				} catch (InvalidMidiDataException e){}
			}		
		}
	}
	
	private static void out(String strMessage)
	{
		System.out.println(strMessage);
	}	

}
