package OJS.Network;
import java.net.*;
import java.io.*;
import java.util.*;

import javax.net.ssl.HttpsURLConnection;
import javax.sound.midi.*;
import java.awt.event.*;
import javax.swing.Timer;
import OJS.Integration.JamClient;
import OJS.Midi.*;

public class Network extends Thread{	
	public static byte MIDI_INFO_PACKET = 1;
	private DatagramSocket socket;
	private DatagramPacket packet;
	private MidiOutputBuffer outputBuffer;
	private MidiInputBuffer inputBuffer;
	private MidiIoController midiController; //TEMP SOLUTION FOR OUTPUTING RECIEVED MIDI/ TODO remove
	boolean activeReceive = false;
	int port = 4445; //base port number.
	int peerPort;
	private String serverURL = "http://abstract.cs.washington.edu/~yanyin/403/server.cgi";
	public String IP;
	public String peerName;
	public InetAddress peer;
	public String information = "";
	
	/**
	 * Initializes the network class by creating a socket connection and collecting the 
	 * basic information about this connection (Host name, port number, ip address).
	 */	
	public Network(){	
		
		//REMOVE TODO <andy>
		//</andy>
		boolean assigned = false; //connect to an available port
		while(!assigned){
			try{
				socket = new DatagramSocket(port);
				port =socket.getLocalPort();
				IP = InetAddress.getLocalHost().getHostAddress();
				assigned = true;
			}
			catch(Exception e){
				port++;
			}
		}
		
		/** it looks like the fragment of code below can be refactored.
		 *  at least you can assign port and ip, which are also needed by other function,
		 *  and reuse them 
		 * */		
		//Get basic information about connection
		try{
			InetAddress temp = InetAddress.getLocalHost();
			InetAddress temp2 = socket.getLocalAddress();
			information +="Connected on port: " +port+"\n"
			+"Host Name: " +temp.getHostName() +"\n"
			+"Domain name of host:" +temp.getCanonicalHostName() +"\n"
			+"IP address: " + IP;

		}
		catch(UnknownHostException e){
			System.out.println("UnkownHost Exception!");
		}
		outputBuffer = new MidiOutputBuffer(8,8);
		inputBuffer =  new MidiInputBuffer (8,8,1);
		inputBuffer.start();
	}
	/**
	 * Connects to the server and sends the users information.
	 * Also creates a stream of data to recieve XML data from teh server.
	 *
	 */
	public String[][] sendToServer(int event, String nickname, String instrument, String IMClient, String IMHandle)throws UnsupportedEncodingException, Exception{
		if(nickname == "" || instrument == "" || IMClient == "" || IMHandle == "") throw new IllegalArgumentException("nickname or instrument or IMClient or IMHandle can't be empty");
		String query = "&IP="+IP+"&Port="+port;
		String param1 = "&nickname="+URLEncoder.encode(nickname,"UTF-8")+
								"&instrument="+URLEncoder.encode(instrument, "UTF-8")+
								"&IMClient="+URLEncoder.encode(IMClient, "UTF-8")+
								"&IMHandle="+URLEncoder.encode(IMHandle, "UTF-8");

		switch(event){
			case JamClient.INSERT:
				query = "?"+"action=insert"+query+param1;							
			break;		
			case JamClient.UPDATE:
				query = "?"+"action=update"+query+param1;
			break;		
			case JamClient.DELETE:
				query = "?"+"action=delete"+query;
			break;		
			default:
				throw new Exception("undefined event");			
		}
		URLConnection server = new URL(serverURL+query).openConnection();	
		BufferedReader in = new BufferedReader(new InputStreamReader(server.getInputStream()));
		String inputLine, result = "";
		while ((inputLine = in.readLine()) != null) result += inputLine;
		in.close();
	
       

		if(result.length() == 0){
			 String [][] dump = new String[1][1];
			 dump [0][0] = "-1";
			return dump;
		}
		else if(result.substring(0, 7).equals("<ERROR>")) return null;
		String[] temp_list = result.split("\t");	
		String[][] list = new String[temp_list.length][6];	
		for(int i = 0; i < temp_list.length; i++){
			int last = 0, first = 0;
			for(int j = 0; j < 6; j++){
				first = temp_list[i].indexOf("\"", last+1);
				last = temp_list[i].indexOf("\"", first+1);
				list[i][j] = temp_list[i].substring(first+1, last); 
			}			
		}		
		return list;
	}	
	
	public void disconnect(){
		socket.disconnect();
	}
	
	public boolean connectToPeer(InetAddress address){
		peer = address;
		socket.connect(address,4444);
		System.out.println("Connected?"+socket.isConnected());
		outputBuffer.start();
		return socket.isConnected();	
	
	}
	
	public void connectToPeer(String address, int port){
		try{
			if(!socket.isConnected()){
				peerName = address;
				peer = InetAddress.getByName(address);
				System.out.println(address);
				socket.connect(peer, port);
				peerPort = port;
				System.out.println("Connected?"+socket.isConnected());
				receive();
				outputBuffer.start();
			}
		}
		catch(UnknownHostException e){
		}	
	}
	
	public void send(String data){
		if(socket.isConnected()){
			DatagramPacket temp = new DatagramPacket(data.getBytes(), data.getBytes().length, peer, peerPort);
			try{
				socket.send(temp);
			}
			catch(IOException e){
				System.out.println("IOException while sending");
				return;
			}
		}
	}
	
	public void send(byte[] bytes){
		if(socket.isConnected()){
			DatagramPacket temp = new DatagramPacket(bytes, bytes.length, peer, peerPort);
			try{
				socket.send(temp);
			}
			catch(IOException e){
				System.out.println("IOException while sending");
				return;
			}
		}
	}
	/**
	 * This method starts the recieving thread if it is not already active.
	 */	
	public void receive(){
		if(!activeReceive){
			this.start();
		}
		else
			System.out.println("Thread Already Active");
	}
	
	/**
	 * Threaded method which sits and waits for a packet to be recieved.
	 */
	public void run(){ 
		activeReceive = true;
		packet = new DatagramPacket(new byte[256], 256);
		while(socket.isConnected()){
			try{
				socket.receive(packet);	
				byte [] data = packet.getData();
				if(data[0] == Network.MIDI_INFO_PACKET && data[1] > 0){   ///IF a midi infor packet, handle!
					inputBuffer.enqueue(data);
				}
					
			}
			catch(IOException e){
				System.out.println("IOException while receiving");
				return;
			}
			// display response
			String received = new String(packet.getData());
			
			//Handle Recieved data;
			packet = new DatagramPacket(new byte[256], 256);
		}
		activeReceive = false;
	}
	
	public void sendSnapShot(MidiSnapShot snap){
		if(socket.isConnected()){
			DatagramPacket temp = new DatagramPacket(snap.encode(), snap.encode().length, peer, peerPort);
			try{
				socket.send(temp);
			}
			catch(IOException e){
				System.out.println("IOException while sending");
				return;
			}
		}
	}

	public void bufferMidi(ShortMessage m, long t){
		
		outputBuffer.enqueue(m,t);
	}
	public void setMidi(MidiIoController c){
		midiController = c;
	}
	
	/* **************************INNER CLASS AREA******************
	 * Created on Jul 26, 2005
	 */
	 class MidiOutputBuffer implements ActionListener{
		int type;
		int bufferSlots;
		MidiSnapShot current = new MidiSnapShot();
		int snapTime;
		
		public MidiOutputBuffer(int bufferTime, int snapTime){
			
			if(bufferTime < snapTime)
				bufferSlots = 1;
			else
				bufferSlots = bufferTime / snapTime;
			
			current = new  MidiSnapShot();
			this.snapTime = snapTime;
		}
		
		public void enqueue(ShortMessage m, long t){
			current.add(m, t);
		}
		
		public void start(){
			new Timer(snapTime, this).start();
		}
		
		public void actionPerformed(ActionEvent evt) {
	      	//All events are timer events
	      			sendSnapShot(current);
	        current.setSize(0);
	      		
	      }
		
	}
	 
	 class MidiInputBuffer implements ActionListener{
	 	LinkedList queue;
		int type;
		int BUFFER_SIZE;
		MidiSnapShot []snaps;
		int bufferSlots;
		MidiSnapShot current = new MidiSnapShot();
		int snapTime;
		
		public MidiInputBuffer(int bufferTime, int snapTime, int buf){//TODO REMOVE SNAP TIME...OR NOT
			BUFFER_SIZE = buf;
			if(bufferTime < snapTime)
				bufferSlots = 1;
			else
				bufferSlots = bufferTime / snapTime;
			
			snaps = new  MidiSnapShot[bufferSlots];
			this.snapTime = snapTime;
			queue = new LinkedList();
		}
		
		public void enqueue(byte[] data){
			   queue.add(current.decode(data));
		}
		
		public void start(){
			new Timer(snapTime, this).start();
		}
		
		public void actionPerformed(ActionEvent evt) {
            //Process Snapshot at front of queue
            if(queue.size() > BUFFER_SIZE &&((MidiSnapShot)(queue.getFirst())).getSize() >= 1 ){
            	//System.out.println("Found valid data!");
                current =  (MidiSnapShot)queue.removeFirst(); //TODO Changed this from remove, to remove first
            	
            	for(int i = 0; i < current.messages.length; i++)
            		midiController.play(current.messages[i], -1);
            	
            	//System.out.println("Playing Note: " + current.messages[0].getData1() +" Velocity: "+  current.messages[0].getData2()+
            	//		" Messagetype ");
            	//System.out.println("Message " + current.messages[0].getData1());
            }
            else if(queue.size() > BUFFER_SIZE){
            	queue.removeFirst();   //TODO Changed this from remove to remove first
            	queue.add(new MidiSnapShot());

            }
            else{
            	queue.add(new MidiSnapShot());
            }
            	
	          //TODO Make it so this actaully sends all midi data in packet!
             // Curently only first note is sent (if it exists)
	      }
		
	}
	
}
